package org.processmining.v7.bruteforce;

public class PlugInStatistics {

	private int numUseFul = 0; //places determined useful are added to the candidate list
	private int numUseless = 0;//useless places, e.g. pure selfloops, are not evaluated
	private int numFitting = 0; //places  evaluated to be  fitting
	private int numUnfitting = 0; //places evaluated to be not fitting (not necessarily malfed)
	private int numRedundant = 0; //number places identified to be redundant (definition monotonicity-paper)
	private int numImpPlace = 0; //number of places identified to be implicit
	private int numCutPaths = 0; //number of cut paths due to malfedness

	private long timeCandFind = 0;
	private long timeEval = 0;
	private long timeReplay = 0;
	private long timeRedTest = 0;
	private long timeImpTest = 0;;

	//singleton-pattern
	private PlugInStatistics() {
	}

	private static PlugInStatistics instance = null;

	public static PlugInStatistics instance() {
		if (instance == null) {
			instance = new PlugInStatistics();
		}
		return instance;
	}

	public void resetStatistics() {
		numUseFul = 0;
		numUseless = 0;
		numFitting = 0;
		numUnfitting = 0;
		numImpPlace = 0;
		numRedundant = 0;
		numCutPaths = 0;

		timeEval = 0;
		timeRedTest = 0;
		timeReplay = 0;
		timeCandFind = 0;
		timeImpTest = 0;

	}

	//Increment num methods

	public void incNumUseFul() {
		numUseFul++;
	}

	public void incNumUseless() {
		numUseless++;
	}
	
	public void incNumUnfitting() {
		numUnfitting++;
	}

	public void incNumCutPaths() {
		numCutPaths++;
	}



	//Increment time methods
	
	public void incTimeReplay(final long time) {
		timeReplay = timeReplay + time;
	}

	public void incTimeRedTest(final long time) {
		timeRedTest = timeRedTest + time;
	}

	public void incTimeCandFind(final long time) {
		timeCandFind = timeCandFind + time;
	}

	public void incTimeImpTest(final long time) {
		timeImpTest = timeImpTest + time;
	}

	//G&S

	public int getNumRedundant() {
		return numRedundant;
	}

	public int getNumUseFul() {
		return numUseFul;
	}

	public int getNumFitting() {
		return numFitting;
	}

	public int getNumUnfitting() {
		return numUnfitting;
	}

	public long getTimeRedTest() {
		return timeRedTest;
	}

	public long getTimeReplay() {
		return timeReplay;
	}

	public long getTimeEval() {
		return timeEval;
	}

	public long getTimeCandFind() {
		return timeCandFind;
	}

	public int getNumCutPaths() {
		return numCutPaths;
	}

	public int getNumUseless() {
		return numUseless;
	}

	public int getNumImpPlace() {
		return numImpPlace;
	}

	public long getTimeImpTest() {
		return timeImpTest;
	}

	public void setNumRedundant(int numRedundant) {
		this.numRedundant = numRedundant;
	}

	public void setNumImpPlace(int numImpPlace) {
		this.numImpPlace = numImpPlace;
	}


	public void setNumFitting(int numFitting) {
		this.numFitting = numFitting;
	}

	public void setTimeEval(long timeEval) {
		this.timeEval = timeEval;
	}

}
