package org.processmining.v7.postproc_after_tc;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Random;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.wizard.ListWizard;
import org.processmining.framework.util.ui.wizard.ProMWizardDisplay;
import org.processmining.framework.util.ui.wizard.ProMWizardStep;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;

@Plugin(name = "eST Miner - Standard", parameterLabels = { "Log" }, returnLabels = {
		"Process Model", "Initial Marking", "Final Marking" }, returnTypes = { Petrinet.class, Marking.class,
				Marking.class }, userAccessible = true, help = "tree traversal including cutting off subtrees, random transitions orderings, post-processing after full traversal")
public class MyFirstMinerPlugin {

	@UITopiaVariant(affiliation = "PADS", author = "Lisa Mannel", email = "mannel@pads.rwth-aachen.de")
	@PluginVariant(requiredParameterLabels = { 0 })
	public Object[] discover(UIPluginContext context, XLog inputLog) throws InterruptedException {
		
		XLog log = (XLog) inputLog.clone();
		
		//Set Parameters
		MyParameters parameters;	
		MyMinerWizardStep myMinerWizardStep = new MyMinerWizardStep(log);
		List<ProMWizardStep<MyParameters>> wizStepList = new ArrayList<>();
		wizStepList.add(myMinerWizardStep);
		ListWizard<MyParameters> listWizard = new ListWizard<>(wizStepList);
		parameters = ProMWizardDisplay.show(context, listWizard, new MyParameters());
		
		
		System.out.println("________________________________________________________________________________________");
		//Initialize progress
		context.getProgress().setIndeterminate(true);
		context.getProgress().setCaption("Initializing...");

		//Preprocess the Log
		//		System.out.println("Preprocessing the Log...");
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XAttributeMap startAttributes = factory.createAttributeMap();
		XAttributeMap endAttributes = factory.createAttributeMap();
		startAttributes.put(XConceptExtension.KEY_NAME,
				factory.createAttributeLiteral(XConceptExtension.KEY_NAME, "start", XConceptExtension.instance()));
		endAttributes.put(XConceptExtension.KEY_NAME,
				factory.createAttributeLiteral(XConceptExtension.KEY_NAME, "end", XConceptExtension.instance()));
		final XEvent startActivity = factory.createEvent(startAttributes);
		final XEvent endActivity = factory.createEvent(endAttributes);
		for (XTrace trace : log) {
			trace.add(0, startActivity);
			trace.add(endActivity);
		}

		//compute average_indices and weights
		final HashMap<String, Integer> weights = weighLog(log, parameters.getClassifier());//compute the number of occurences of each transition
		final HashMap<String, Integer> average_indices = measureLog(log, parameters.getClassifier(), weights);//compute the average index for each transition (assuming no duplicates)

		Path p = Paths.get("U:/MyLogs/Log2/Log2-late first.txt"); //name and place of data-storing file

		//compute in transition order
		String[] transitions = computeTransitions(weights); //ordered light activities first
		//String[] transitions = computeTransitions(average_indices); //ordered early activities first
		transitions = shuffleTransitions(transitions.clone());

		//compute out transition order
		//String[] outtransitions=transitions.clone();
		//String[] outtransitions = reverseTransitions(transitions.clone());
		String[] outtransitions = shuffleTransitions(transitions.clone());

		int endIndex = 0;//find end index
		for (int i = 0; i < transitions.length; i++) {
			if (parameters.getClassifier().getClassIdentity(endActivity).equals(transitions[i])) {
				endIndex = i;
				break;
			}
		}
		//move end to position zero
		while (endIndex != 0) {
			String temp = transitions[endIndex - 1];
			transitions[endIndex - 1] = transitions[endIndex];
			transitions[endIndex] = temp;
			endIndex--;
		}
		int startIndex = 0;//find start index
		for (int i = 0; i < transitions.length; i++) {
			if (parameters.getClassifier().getClassIdentity(startActivity).equals(transitions[i])) {
				startIndex = i;
				break;
			}
		}
		int OutStartIndex = 0;//find start index
		for (int i = 0; i < outtransitions.length; i++) {
			if (parameters.getClassifier().getClassIdentity(startActivity).equals(outtransitions[i])) {
				OutStartIndex = i;
				break;
			}
		}
		//compute out transition mapping with position zero mapping to start
		int[] outTransitionsMapping = computeOutTransitionMapping(transitions.clone(), OutStartIndex, outtransitions);

		String inorder = "Inorder: ";
		String outorder = "Outorder: ";
		for (int i = 0; i < transitions.length; i++) {
			inorder = inorder + transitions[i] + "; ";
			outorder = outorder + transitions[outTransitionsMapping[i]] + "; ";
		}
		System.out.println(inorder);
		System.out.println(outorder);
		System.out.println("---------------------------------------------------------");

		//more preprocessing
		convertWeights(weights, transitions);
		convertDistances(average_indices, transitions);
		final ArrayList<ArrayList<Integer>> convertedLog = convertLog(log, transitions, parameters.getClassifier());
		final Object[] groupedLog = groupLog(convertedLog);
		final ArrayList<ArrayList<Integer>> uniqueTraces = (ArrayList<ArrayList<Integer>>) groupedLog[0];
		final HashMap<ArrayList<Integer>, Integer> traceCounts = (HashMap<ArrayList<Integer>, Integer>) groupedLog[1];
		System.out.println("Number of Traces: " + convertedLog.size() + ", Unique Variants: " + uniqueTraces.size());

		//Initialize Process Model (unique transitions and empty list of places)
		System.out.println("Number of Event Classes: " + transitions.length);

		ArrayList<MyPlace> places = new ArrayList<MyPlace>();
		MyProcessModel pM = new MyProcessModel(places, transitions);

		//---------------------Adding Places (the important part)---------------------------------------------------------------------
		//Adjust progress
		//		context.getProgress().setCaption("Evaluate and add places...");

		//Initialize Candidate Traversal based on Transitions[]
		System.out.println("--------------------------------------------------------------------------");
		System.out.println("Start adding places...");
		MyCandidates candidates = new MyCandidates(transitions, startIndex, endIndex, outTransitionsMapping,
				parameters.getPlace_restrictions());
		//Evaluating and possibly adding the candidate places
		MyPlaceEvaluator evaluator = new MyPlaceEvaluator(pM, uniqueTraces, traceCounts, parameters.getThreshold(), parameters.isSafeness());
		ProcessDiscovery pDiscovery = new ProcessDiscovery(pM, candidates, evaluator);
		final long timeEvalStart = System.currentTimeMillis();
		pDiscovery.start();
		pDiscovery.join(parameters.getTimeAllowance());
		if (pDiscovery.isAlive()) {
			pDiscovery.interrupt();
		}
		PlugInStatistics.instance().setTimeEval(System.currentTimeMillis() - timeEvalStart);
		MyProcessModel finalPM = new MyProcessModel(new ArrayList<MyPlace>(pDiscovery.getpM().getPlaces()),
				pDiscovery.getpM().getTransitions());
		PlugInStatistics.instance().setNumFitting(PlugInStatistics.instance().getNumUseFul()-PlugInStatistics.instance().getNumUnfitting());
		
		//---------------------------Preliminary Results----------------------------------------------------
		
		//compute results
		int numCurrentPlaces = finalPM.getPlaces().size();
		final int P_possible = (int) Math.pow(Math.pow(2, pM.getTransitions().length), 2);
		final int P_all = (int) (Math.pow(Math.pow(2, (pM.getTransitions().length-1)), 2)-2*Math.pow(2, (pM.getTransitions().length-1))+1);
		final int cutoffP = (P_all - PlugInStatistics.instance().getNumUseFul()
				- PlugInStatistics.instance().getNumUseless());
		final int cutoffPPercentage = (int) ((cutoffP * 100.0) / P_possible);
		final int cutoffPPercentageP_all = (int) ((cutoffP * 100.0) / P_all);

		//Record the results to text file
		String transitionsInOrder = "";
		String transitionsWeights = "";
		String transitionsIndices = "";
		for (String s : transitions) {
			transitionsInOrder = transitionsInOrder + s + ", ";
			transitionsWeights = transitionsWeights + weights.get(s) + ", ";
			transitionsIndices = transitionsIndices + average_indices.get(s) + ", ";
		}
		Charset charset = Charset.forName("US-ASCII");
		String outputToFile = convertedLog.size() + "\r\n" + uniqueTraces.size() + "\r\n" + transitions.length + "\r\n"
				+ (int) Math.pow(Math.pow(2, pM.getTransitions().length), 2) + "\r\n"
				+ PlugInStatistics.instance().getNumUseFul() + "\r\n" + PlugInStatistics.instance().getNumUseless()
				+ "\r\n" + PlugInStatistics.instance().getNumUnfitting() + "\r\n"
				+ PlugInStatistics.instance().getNumCutPaths() + "\r\n" + cutoffP + "\r\n" + cutoffPPercentage + "\r\n"
				+ PlugInStatistics.instance().getNumFitting() + "\r\n" + PlugInStatistics.instance().getNumRedundant()
				+ "\r\n" + PlugInStatistics.instance().getNumImpPlace() + "\r\n" + numCurrentPlaces + "\r\n"
				+ transitionsInOrder + "\r\n" + transitionsWeights + "\r\n" + transitionsIndices + "\r\n";

		try (BufferedWriter writer = Files.newBufferedWriter(p, charset)) {
			writer.write(outputToFile, 0, outputToFile.length());
		} catch (IOException x) {
			System.err.format("IOException writing File: %s%n", x);
		}

		//print results for debugging
		System.out.println("_____________________________________________________________________________________________");
		System.out.println("Number of possible Places: " + P_possible);
		System.out.println("P-all: " + P_all);
		System.out.println("Number of identified useful Places: " + PlugInStatistics.instance().getNumUseFul());
		System.out.println("Number of identified useless places: " + PlugInStatistics.instance().getNumUseless());
		System.out.println("Number of replayed unfitting Places: " + PlugInStatistics.instance().getNumUnfitting());
		System.out.println("Cutoff Places: " + cutoffP + " | " + cutoffPPercentage + " % of P-possible | "
				+ cutoffPPercentageP_all + " % of P-all");
		System.out.println("P-fit: " + PlugInStatistics.instance().getNumFitting());
		System.out.println("------------------------------------------------------------------------------------------------------");
		System.out.println("Evaluating Places needed: " + (PlugInStatistics.instance().getTimeEval()) + " ms");
		System.out
				.println("of this for finding candidates: " + (PlugInStatistics.instance().getTimeCandFind()) + " ms");
		System.out.println("Of this for Replay: " + PlugInStatistics.instance().getTimeReplay() + " ms");
		System.out.println("-----------------------------------------------------------------------------------------------------");


		//--------------------------Post-Processing---------------------------------------------------------------------------------
		//Adjust progress
		context.getProgress().setCaption("Post-processing...");
		System.out.println("Start post-processing...");
		final long postprocstart = System.currentTimeMillis();

		//remove first round of redundant places TODO: compare with correpnsing prom plugin, check for bugs (I remove to many places?)
		//test for redundancy as defined in the monotonicity paper
		RedundancyTester redundancyTester = new RedundancyTester();
		System.out.println("P-fit: " + numCurrentPlaces);
		finalPM = redundancyTester.testForRedundancy(finalPM, 25);
		PlugInStatistics.instance().setNumRedundant(numCurrentPlaces - finalPM.getPlaces().size());

		//remove implicit places
		numCurrentPlaces = finalPM.getPlaces().size();
		ImplicitnessTester implicitnessTester = new ImplicitnessTester();
//		System.out.println("Number of places before removing implicit places (LPP2): " + numCurrentPlaces);
		finalPM = implicitnessTester.testByLPP2(finalPM);
		PlugInStatistics.instance().setNumImpPlace(numCurrentPlaces - finalPM.getPlaces().size());

		//remove remaining redundant places
		numCurrentPlaces = finalPM.getPlaces().size();
//		System.out.println("Number of places before removing redundant places (2): " + finalPM.getPlaces().size());
		finalPM = redundancyTester.testForRedundancy(finalPM, 60);
		PlugInStatistics.instance().setNumRedundant(
				PlugInStatistics.instance().getNumRedundant() + (numCurrentPlaces - finalPM.getPlaces().size()));
		numCurrentPlaces = finalPM.getPlaces().size();

		//remove remaining implicit places by old method
		numCurrentPlaces = finalPM.getPlaces().size();
//		System.out.println("Number of places before removing implicit places (2): " + finalPM.getPlaces().size());
		finalPM = implicitnessTester.testForImplicitPlaces(finalPM);
		numCurrentPlaces = finalPM.getPlaces().size();

		Long postproctime=System.currentTimeMillis()-postprocstart;
		System.out.println("P-final: " + numCurrentPlaces);

		//print for debugging
		System.out.println("Post-processing needed: " + postproctime + " ms");
		System.out.println("-------------------------------------------------------------------------------------");

		//-----------------create Petrinet from Process Model-------------------------------------------------------------------------
		Petrinet net = PetrinetFactory.newPetrinet("Process Model");
		Marking initial_marking = new Marking();
		Marking final_marking = new Marking();
		//add start place and add it to to initial markings
		Place startP = net.addPlace("Start");
		initial_marking.add(startP);
		//add end place and add it to final markings
		Place endP = net.addPlace("End");
		final_marking.add(endP);
		//add transitions and connecting arcs for start and end place
		for (int i = 0; i < finalPM.getTransitions().length; i++) {//add the transitions to the petri net
			Transition t = net.addTransition(finalPM.getTransitions()[i]);
			if (i == startIndex) {
				net.addArc(startP, t);
			} else if (i == endIndex) {
				net.addArc(t, endP);
			}
		}

		//add intermediate places from process model
		final Collection<Transition> petriTransitions = net.getTransitions();
		for (MyPlace myP : finalPM.getPlaces()) {
			Place newP = net.addPlace("");
			for (Transition t : petriTransitions) {
				if (getTransitionNames(myP.getInputTrKey(), transitions).contains(t.getLabel())) {
					net.addArc(t, newP);
				}
				if (getTransitionNames(myP.getOutputTrKey(), transitions).contains(t.getLabel())) {
					net.addArc(newP, t);
				}
			}
		}
		context.getConnectionManager().addConnection(new InitialMarkingConnection(net, initial_marking));
		context.getConnectionManager().addConnection(new FinalMarkingConnection(net, final_marking));

		System.out.println("Number of Arcs: " + net.getEdges().size());
		System.out.println("____________________________________________________________________________________");

		PlugInStatistics.instance().resetStatistics();
		return new Object[] { net, initial_marking, final_marking };
	}

	//_____________________________END OF MAIN__________________________________________________________________________________________

	//make sure that mapping to start is at position 0
	private int[] computeOutTransitionMapping(String[] intransitions, int startIndex, String[] outtransitions) {
		for (int i = startIndex; i > 0; i--) {//move start to posiotion 0
			String temp = outtransitions[i - 1];
			outtransitions[i - 1] = outtransitions[i];
			outtransitions[i] = temp;
		}
		//compute mapping
		int[] outTransitionMapping = new int[outtransitions.length];
		for (int i = 0; i < intransitions.length; i++) {
			for (int o = 0; o < outtransitions.length; o++) {
				if (intransitions[i] == outtransitions[o]) {
					outTransitionMapping[o] = i;
					break;
				}
			}
		}
		return outTransitionMapping;
	}

	//for a given log returns a log containing all trace variants and a map containing the number of their occurences
	private Object[] groupLog(final ArrayList<ArrayList<Integer>> givenLog) {
		HashMap<ArrayList<Integer>, Integer> numTraces = new HashMap<ArrayList<Integer>, Integer>();
		ArrayList<ArrayList<Integer>> log = new ArrayList<ArrayList<Integer>>();
		for (ArrayList<Integer> trace : givenLog) {
			if (!numTraces.containsKey(trace)) {
				numTraces.put(trace, 0);
			}
			numTraces.put(trace, numTraces.get(trace) + 1);
		}
		//		System.out.println(numTraces);
		for (ArrayList<Integer> variant : numTraces.keySet()) {
			log.add(variant);
		}
		return new Object[] { log, numTraces };
	}

	//replace activities by integers corresponding to their position in the transitions array
	private ArrayList<ArrayList<Integer>> convertLog(XLog log, String[] transitions, XEventClassifier classifier) {
		ArrayList<ArrayList<Integer>> newLog = new ArrayList<ArrayList<Integer>>();
		for (XTrace trace : log) {
			ArrayList<Integer> newTrace = new ArrayList<Integer>();
			for (XEvent event : trace) {
				String transitionName = classifier.getClassIdentity(event);
				for (int i = 0; i < transitions.length; i++) {
					if (transitions[i].equals(transitionName)) {
						newTrace.add(i);
						break;
					}
				}
			}
			newLog.add(newTrace);
		}
		return newLog;
	}

	//replaces the string key by the integer corresponding to the position of this transition in the transitions array
	private HashMap<Integer, Integer> convertWeights(final HashMap<String, Integer> weights,
			final String[] transitions) {
		HashMap<Integer, Integer> result = new HashMap<Integer, Integer>();
		for (int i = 0; i < transitions.length; i++) {
			int weight = weights.get(transitions[i]);
			result.put(i, weight);
		}
		return result;
	}

	//replaces the string key by the integer corresponding to the position of this transition in the transitions array
	private HashMap<Integer, Integer> convertDistances(final HashMap<String, Integer> average_indices,
			final String[] transitions) {
		HashMap<Integer, Integer> result = new HashMap<Integer, Integer>();
		for (int i = 0; i < transitions.length; i++) {
			int avIndex = average_indices.get(transitions[i]);
			result.put(i, avIndex);
		}
		return result;
	}

	//create transition array ordered according to weights
	private String[] computeTransitions(HashMap<String, Integer> weights) {
		Collection<String> weighedEvents = weights.keySet();
		ArrayList<String> transitionList = new ArrayList<String>();
		ArrayList<Integer> tempWeights = new ArrayList<Integer>(weights.values());
		tempWeights.sort(null);
		java.util.Iterator<Integer> iterator = tempWeights.iterator();
		while (iterator.hasNext()) {
			int value = iterator.next();
			for (String eventName : weighedEvents) {
				if (weights.get(eventName) == value) { //the key "activity" is mapped to the value
					transitionList.add(eventName);
				}
			}
			while (!tempWeights.isEmpty() && (tempWeights.get(0) == value)) {//ensure the transitions mapped to from the current weight are added only once
				tempWeights.remove(0);
			}
			iterator = tempWeights.iterator();
		}
		String[] transitions = new String[transitionList.size()];
		for (int i = 0; i < transitions.length; i++) {
			transitions[i] = transitionList.get(i);
		}
		return transitions;
	}

	//compute weight of events (number of occurences in the whole log)
	private HashMap<String, Integer> weighLog(XLog log, XEventClassifier classifier) {
		HashMap<String, Integer> weights = new HashMap<String, Integer>();
		for (XTrace trace : log) {
			for (XEvent event : trace) {
				String key = classifier.getClassIdentity(event);
				if (weights.containsKey(key)) {
					int weight = weights.get(key);
					weights.put(key, weight + 1);
				} else {
					weights.put(key, 1);
				}
			}
		}
		return weights;
	}

	//returns the average indices of the activities. prone to errors when duplicates exist
	private HashMap<String, Integer> measureLog(XLog log, XEventClassifier classifier,
			HashMap<String, Integer> weights) {
		HashMap<String, Integer> average_indices = new HashMap<String, Integer>();
		//compute the sum of all indices
		for (XTrace trace : log) {
			for (int i = 0; i < trace.size(); i++) {
				XEvent event = trace.get(i);
				String key = classifier.getClassIdentity(event);
				if (average_indices.containsKey(key)) {
					int index = average_indices.get(key);
					average_indices.put(key, index + i);
				} else {
					average_indices.put(key, i);
				}
			}
		}
		//use the weights (=number of occurences) to compute the average indices
		for (String key : average_indices.keySet()) {
			int indexSum = average_indices.get(key);
			indexSum = indexSum / weights.get(key);
			average_indices.put(key, indexSum);
		}
		return average_indices;
	}

	//randomly shuffles the transitions array TODO check whether this really works as intended
	private String[] shuffleTransitions(String[] transitions) {
		int index;
		String temp;
		Random random = new Random();
		for (int i = transitions.length - 1; i > 0; i--) {
			index = random.nextInt(i + 1);
			temp = transitions[index];
			transitions[index] = transitions[i];
			transitions[i] = temp;
		}
		return transitions;
	}

	//returns the given array in reversed order
	private String[] reverseTransitions(String[] transitions) {
		String[] reversed_transitions = new String[transitions.length];
		for (int i = 0; i < transitions.length; i++) {
			reversed_transitions[transitions.length - 1 - i] = transitions[i];
		}
		return reversed_transitions;
	}

	//returns a collection containing all transitions names from the given transitions array
	private Collection<String> getTransitionNames(final int key, final String[] transitions) {
		Collection<String> result = new ArrayList<String>();
		if (key > (Math.pow(2, transitions.length))) {
			return null;
		}
		for (int i = 0; i < transitions.length; i++) {
			if ((key & getMask(i, transitions)) > 0) { //test key for ones
				result.add(transitions[i]);
			}
		}
		return result;
	}

	//for a given position in the transition array return the corresponding bitmask
	private int getMask(final int pos, final String[] transitions) {
		return 1 << (transitions.length - 1 - pos);
	}

}