package org.processmining.v7.postproc_during_tc;


public class MyPlace implements Comparable<MyPlace> {
	private final int inputTrKey;
	private final int outputTrKey;

	private int currentTokens;
	private int underfed;
	private boolean activated;

	public MyPlace(final int inputTrKey, final int outputTrKey) {
		this.inputTrKey = inputTrKey;
		this.outputTrKey = outputTrKey;
		currentTokens = 0;
		underfed = 0;
		activated = false;
	}

	public MyPlace() {
		inputTrKey = 0;
		outputTrKey = 0;
		currentTokens = 0;
		underfed = 0;
		activated = false;
	}

	public MyPlace copy() {
		int newUf = new Integer(getUnderfed());
		int newTokens = new Integer(getCurrentTokens());
		int newInKey = new Integer(getInputTrKey());
		int newOutKey = new Integer(getOutputTrKey());
		MyPlace newP = new MyPlace(newInKey, newOutKey);
		newP.setCurrentTokens(newTokens);
		newP.setUnderfed(newUf);
		if (getActivated()) {
			newP.setActivated(true);
		}
		return newP;
	}

	public String toString() {
		return "(" + getInputTrKey() + "|" + getOutputTrKey() + ")";
	}

	public String toBinaryString() {
		return "(" + Long.toBinaryString(getInputTrKey()) + "|" + Long.toBinaryString(getOutputTrKey()) + ")";
	}


	/*	public boolean equals(final MyPlace p) {
		if ((getInputTrKey() == p.getInputTrKey()) && (getOutputTrKey() == p.getOutputTrKey())) {
			return true;
		}
		return false;
	}

/*	//implementing comparable --- important for ordering! --> true ordering, below only number of transitions
	public int compareTo(final MyPlace p) {
		long in1 = getInputTrKey();
		long in2 = p.getInputTrKey();
		long out1 = getOutputTrKey();
		long out2 = p.getOutputTrKey();

		if ((Long.bitCount(in1) + Long.bitCount(out1)) < (Long.bitCount(in2) + Long.bitCount(out2))) {//p1 is higher level (less transitions)
			return -1;
		}
		if ((Long.bitCount(in1) + Long.bitCount(out1)) > (Long.bitCount(in2) + Long.bitCount(out2))) {//p1 is lower level (more transitions)
			return 1;
		}
		//compare on the same level
		if (Long.bitCount(in1) < Long.bitCount(in2)) {
			return -1;
		}
		if (Long.bitCount(in1) > Long.bitCount(in2)) {
			return 1;
		}
		//same set sizes
		if (in1 > in2) {//bigger value means the transitions of smaller order is included
			return -1;
		}
		if (in1 < in2) {
			return 1;
		}
		//in1=in2
		if (out1 > out2) {
			return -1;
		}
		if (out1 < out2) {
			return 1;
		}
		return 0;
	}
	*/
	
	//implementing comparable --- important for ordering! order many-transition-places first
	public int compareTo(final MyPlace p) {
		int in1 = this.getInputTrKey();
		int in2 = p.getInputTrKey();
		int out1 = this.getOutputTrKey();
		int out2 = p.getOutputTrKey();

		if (Integer.bitCount(in1)+Integer.bitCount(out1) > Integer.bitCount(in2)+Integer.bitCount(out2)) {
			return 1;
		}
		else if(Integer.bitCount(out1) > Integer.bitCount(out2)) {
			return 1;
		}
		if (Integer.bitCount(in1)+Integer.bitCount(out1) < Integer.bitCount(in2)+Integer.bitCount(out2)) {
			return -1;
		}
		else if(Integer.bitCount(out1) < Integer.bitCount(out2)) {
			return -1;
		}		
		return 0;
	}
	
	public boolean equals(final MyPlace p) {
		if (this.compareTo(p) == 0) {
			return true;
		}
		return false;
	}
	

	public void addToken() {
		currentTokens++;
	}

	public void consumeToken() {
		if (currentTokens == 0) {
			underfed++;
		} else {
			currentTokens--;
		}
	}

	public void activate() {
		activated = true;
	}

	public void fire(final int mask) {
		if ((getOutputTrKey() & mask) > 0) {//if place contains transition indicated by mask as output
			consumeToken();
			activate();
		}
		if ((getInputTrKey() & mask) > 0) {//if place contains transition indicated by mask as input
			addToken();
			activate();
		}
	}

	//G&S
	public int getInputTrKey() {
		return inputTrKey;
	}

	public int getOutputTrKey() {
		return outputTrKey;
	}

	public int getCurrentTokens() {
		return currentTokens;
	}

	public int getUnderfed() {
		return underfed;
	}

	public void setCurrentTokens(final int currentTokens) {
		this.currentTokens = currentTokens;
	}

	public void setUnderfed(final int underfed) {
		this.underfed = underfed;
	}

	public boolean getActivated() {
		return activated;
	}

	public void setActivated(final boolean activated) {
		this.activated = activated;
	}



}
