package org.processmining.v7.postproc_during_tc;

import java.util.ArrayList;
import java.util.HashMap;

public class MyPlaceEvaluator {

	private final ArrayList<ArrayList<Integer>> log;
	private final double threshold;
	private boolean safeness; 
	private final MyProcessModel pM;
	private final HashMap<ArrayList<Integer>, Integer> traceCounts;

	public MyPlaceEvaluator(final MyProcessModel pM, final ArrayList<ArrayList<Integer>> singleTraceLog,
			final HashMap<ArrayList<Integer>, Integer> traceCounts, final double threshold, final boolean safeness) {
		this.pM = pM;
		this.threshold = threshold;
		log = singleTraceLog;
		this.traceCounts = traceCounts;
		this.safeness= safeness;
	}

	public PlaceStatus testPlace(final MyPlace current) {
		long timeReplayStart = System.currentTimeMillis();
		PlaceStatus result = replayFull(current);
		PlugInStatistics.instance().incTimeReplay(System.currentTimeMillis() - timeReplayStart);
		return result;
	}

	private PlaceStatus replayFull(final MyPlace current) {
		double numUf = 0; //number of underfed traces
		double numOf = 0; //number of overfed traces
		double numActTr = 0;
		double numFitAct = 0;
		double numFit = 0;
		double numUnsafe = 0;
//		final int numTr = log.size(); //number of all traces

		//replay each trace on the place
		for (ArrayList<Integer> trace : log) {
			boolean safeTrace = true;
			MyPlace currentCopy = current.copy();
			for (int event : trace) {
				currentCopy.fire(getMask(event));
				if(this.safeness && currentCopy.getCurrentTokens()>1) {
					safeTrace=false;
				}
			}
			//save results of replay of this trace
			if(!safeTrace) {
				numUnsafe=numUnsafe+traceCounts.get(trace);
			}
			if (currentCopy.getActivated()) {
				numActTr = numActTr + traceCounts.get(trace);
				if (currentCopy.getCurrentTokens() > 0) {
					numOf = numOf + traceCounts.get(trace);
				}
				if (currentCopy.getUnderfed() > 0) {
					numUf = numUf + traceCounts.get(trace);
				}
				if ((currentCopy.getUnderfed() == 0) && (currentCopy.getCurrentTokens() == 0)) {
					numFitAct = numFitAct + traceCounts.get(trace);
					numFit = numFit + traceCounts.get(trace);
				}
			} else {//non activated cannot be malfed
				numFit = numFit + traceCounts.get(trace);
			}
		}
		//use results of replay
		if (numActTr == 0 || numUnsafe>(1-threshold)) {//place not activated or unsafe( and safeness-check activated)
			PlugInStatistics.instance().incNumUnfitting();
			return PlaceStatus.UNFIT;
		}
		double relFit = numFitAct / numActTr; //fraction of fitting activating traces
		//double freqFit = (double) numFit / (double) numTr; //fraction of fitting traces
		if ((relFit >= threshold)) {//enough evidence that place is fit
			return PlaceStatus.FIT;
		} else {
			PlugInStatistics.instance().incNumUnfitting();
			PlaceStatus result = PlaceStatus.UNFIT;
			double relUf = numUf / numActTr;
			if (relUf > (1 - threshold)) { //enough evidence that place is underfed
				result = PlaceStatus.UNDERFED;
			}
			double relOf = numOf / numActTr;
			if (relOf > (1 - threshold)) { //enough evidence that place is overfed
				if (result == PlaceStatus.UNDERFED) {
					result = PlaceStatus.MALFED;
				} else {
					result = PlaceStatus.OVERFED;
				}
			}
			return result;
		}
	}

	private int getMask(final int position) {
		return (1 << (pM.getTransitions().length - 1 - position));
	}

}
